#!/usr/bin/env python3
# pylint: disable=unused-argument
import sys
import os

import mmpm.utils
import mmpm.core
import mmpm.opts
import mmpm.consts
import mmpm.models

from typing import List, Dict


MagicMirrorPackage = mmpm.models.MagicMirrorPackage
get_env = mmpm.utils.get_env

__version__ = 2.05


def main(argv):
    ''' Main entry point for CLI '''

    parser = mmpm.opts.get_user_args()
    args, additional_args = parser.parse_known_args()

    if args.version:
        print(f'{__version__}')
        sys.exit(0)

    if args.migrate:
        mmpm.core.migrate()
        sys.exit(0)

    mmpm.utils.assert_required_defaults_exist()

    if args.guided_setup:
        mmpm.core.guided_setup()
        sys.exit(0)

    if args.subcmd in mmpm.opts.SINGLE_OPTION_ARGS and not mmpm.utils.assert_one_option_selected(args):
        mmpm.utils.fatal_too_many_options(args)

    creation_date, expiration_date = mmpm.utils.calculate_expiration_date_of_database()
    databased_expired: bool = mmpm.utils.should_refresh_database(creation_date, expiration_date)
    should_refresh: bool = True if args.subcmd == mmpm.opts.DATABASE and args.refresh else databased_expired

    packages: Dict[str, List[MagicMirrorPackage]] = mmpm.core.load_packages(force_refresh=should_refresh)

    if databased_expired and args.subcmd != mmpm.opts.UPDATE and mmpm.core.check_for_mmpm_updates(automated=True):
        print('Upgrade available for MMPM. Execute `pip3 install --upgrade --user mmpm` to perform the upgrade now')

    if not packages:
        mmpm.utils.fatal_msg('Unable to retrieve packages. Please check your internet connection.')

    if args.subcmd == mmpm.opts.LIST:
        if args.installed:
            installed_packages = mmpm.core.get_installed_packages(packages)

            if not installed_packages:
                mmpm.utils.fatal_msg('No packages are currently installed')

            mmpm.core.display_packages(installed_packages, title_only=args.title_only, include_path=True)

        elif args.all:
            mmpm.core.display_packages(packages, title_only=args.title_only)
        elif args.exclude_local:
            excluded = mmpm.utils.get_difference_of_packages(packages, mmpm.core.get_installed_packages(packages))
            mmpm.core.display_packages(excluded, title_only=args.title_only)
        elif args.categories:
            mmpm.core.display_categories(packages, title_only=args.title_only)
        elif args.gui_url:
            print(mmpm.core.get_web_interface_url())
        elif args.upgradeable:
            mmpm.core.display_available_upgrades()
        else:
            mmpm.utils.fatal_no_arguments_provided(args.subcmd)

    elif args.subcmd == mmpm.opts.SHOW:
        if not additional_args:
            mmpm.utils.fatal_no_arguments_provided(args.subcmd)

        if args.remote:
            health: dict = mmpm.utils.get_remote_repo_api_health()

            for api in health:
                if health[api][mmpm.consts.ERROR]:
                    mmpm.utils.fatal_msg(health[api][mmpm.consts.ERROR])
                elif health[api][mmpm.consts.WARNING]:
                    mmpm.utils.warning_msg(health[api][mmpm.consts.WARNING])

        for query in additional_args:
            result = mmpm.core.search_packages(packages, query, by_title_only=True)

            if not result:
                mmpm.utils.fatal_msg(f'Unable to match {query} to a package title')

            mmpm.core.show_package_details(result, args.remote)

    elif args.subcmd == mmpm.opts.OPEN:
        import webbrowser

        if additional_args:
            mmpm.utils.fatal_invalid_additional_arguments(args.subcmd)
        elif args.config:
            mmpm.utils.open_default_editor(os.path.join(get_env(mmpm.consts.MMPM_MAGICMIRROR_ROOT_ENV), 'config', 'config.js'))
        elif args.custom_css:
            mmpm.utils.open_default_editor(os.path.join(get_env(mmpm.consts.MMPM_MAGICMIRROR_ROOT_ENV), 'custom', 'custom.css'))
        elif args.magicmirror:
            webbrowser.open(get_env(mmpm.consts.MMPM_MAGICMIRROR_URI_ENV))
        elif args.gui:
            webbrowser.open(mmpm.core.get_web_interface_url())
        elif args.mm_wiki:
            webbrowser.open(mmpm.consts.MAGICMIRROR_WIKI_URL)
        elif args.mm_docs:
            webbrowser.open(mmpm.consts.MAGICMIRROR_DOCUMENTATION_URL)
        elif args.mmpm_wiki:
            webbrowser.open(mmpm.consts.MMPM_WIKI_URL)
        elif args.mmpm_env:
            mmpm.utils.open_default_editor(mmpm.consts.MMPM_ENV_FILE)
        else:
            mmpm.utils.fatal_no_arguments_provided(args.subcmd)

    elif args.subcmd == mmpm.opts.ADD_EXT_PKG:
        if args.remove:
            mmpm.core.remove_external_package_source(
                [mmpm.utils.sanitize_name(package) for package in args.remove],
                assume_yes=args.assume_yes
            )
        else:
            mmpm.core.add_external_package(args.title, args.author, args.repo, args.desc)

    elif args.subcmd == mmpm.opts.UPDATE:
        if additional_args:
            mmpm.utils.fatal_invalid_additional_arguments(args.subcmd)

        package_upgrades: List[MagicMirrorPackage] = mmpm.core.check_for_package_updates(packages)
        magicmirror_upgrade: bool = mmpm.core.check_for_magicmirror_updates()
        mmpm_upgrade: bool = mmpm.core.check_for_mmpm_updates()
        total = len(package_upgrades) + int(mmpm_upgrade) + int(magicmirror_upgrade)

        if not total:
            print('All packages are up to date')
        else:
            message: str = f"{total} {'upgrade' if total == 1 else 'upgrades'} {'is' if total == 1 else 'are'} available"
            print(f'{message}. Run `mmpm list --upgradeable` for details')

    elif args.subcmd == mmpm.opts.UPGRADE:
        mmpm.core.upgrade_available_packages_and_applications(args.assume_yes, additional_args)

    elif args.subcmd == mmpm.opts.INSTALL:
        if args.gui:
            mmpm.core.install_mmpm_gui()
        elif args.as_module:
            mmpm.core.install_mmpm_as_magicmirror_module()
        elif args.magicmirror:
            mmpm.core.install_magicmirror()
        elif args.autocomplete:
            mmpm.core.install_autocompletion(assume_yes=args.assume_yes)
        elif not additional_args:
            mmpm.utils.fatal_no_arguments_provided(args.subcmd)
        else:
            installation_candidates = mmpm.core.get_installation_candidates(
                packages,
                [mmpm.utils.sanitize_name(package) for package in additional_args]
            )

            mmpm.core.install_packages(installation_candidates, assume_yes=args.assume_yes)

    elif args.subcmd == mmpm.opts.REMOVE:
        if args.gui:
            mmpm.core.remove_mmpm_gui()

        elif not additional_args:
            mmpm.utils.fatal_no_arguments_provided(args.subcmd)

        installed_packages = mmpm.core.get_installed_packages(packages)

        if not installed_packages:
            mmpm.utils.fatal_msg("No packages are currently installed")

        mmpm.core.remove_packages(
            installed_packages,
            [mmpm.utils.sanitize_name(package) for package in additional_args],
            assume_yes=args.assume_yes
        )

    elif args.subcmd == mmpm.opts.SEARCH:
        if not additional_args:
            mmpm.utils.fatal_no_arguments_provided(args.subcmd)

        if len(additional_args) > 1:
            mmpm.utils.fatal_msg(f'Too many arguments. `mmpm {args.subcmd}` only accepts one search argument')
        else:
            if args.exclude_local:
                packages = mmpm.utils.get_difference_of_packages(packages, mmpm.core.get_installed_packages(packages))

            mmpm.core.display_packages(
                mmpm.core.search_packages(packages, additional_args[0], case_sensitive=args.case_sensitive),
                title_only=args.title_only
            )

    elif args.subcmd == mmpm.opts.MM_CTL:
        if additional_args:
            mmpm.utils.fatal_invalid_additional_arguments(args.subcmd)
        elif args.status:
            mmpm.core.display_magicmirror_modules_status()
        elif args.hide:
            mmpm.core.hide_magicmirror_modules(args.hide)
        elif args.show:
            mmpm.core.show_magicmirror_modules(args.show)
        elif args.start or args.stop or args.restart or args.rotate != None:
            MMPM_IS_DOCKER_IMAGE: str = mmpm.utils.get_env('MMPM_IS_DOCKER_IMAGE')
            if MMPM_IS_DOCKER_IMAGE:
                mmpm.utils.fatal_msg('Cannot execute this command within a docker image')
            elif args.start:
                mmpm.core.start_magicmirror()
            elif args.stop:
                mmpm.core.stop_magicmirror()
            elif args.restart:
                mmpm.core.restart_magicmirror()
            elif args.rotate != None:
                mmpm.core.rotate_raspberrypi_screen(args.rotate)
        else:
            mmpm.utils.fatal_no_arguments_provided(args.subcmd)

    elif args.subcmd == mmpm.opts.DATABASE:
        if args.refresh:
            sys.exit(0)
        if additional_args:
            mmpm.utils.fatal_invalid_additional_arguments(args.subcmd)
        elif args.details:
            mmpm.core.database_details(packages)
        elif args.dump:
            mmpm.core.dump_database()
        else:
            mmpm.utils.fatal_no_arguments_provided(args.subcmd)

    elif args.subcmd == mmpm.opts.LOG:
        if additional_args:
            mmpm.utils.fatal_invalid_additional_arguments(args.subcmd)
        elif args.zip:
            mmpm.core.zip_mmpm_log_files()
        elif not args.cli and not args.gui:
            # if the user doesn't provide arguments, just display everything, but consider the --tail arg
            mmpm.core.display_log_files(True, True, args.tail)
        else:
            mmpm.core.display_log_files(args.cli, args.gui, args.tail)

    elif args.subcmd == mmpm.opts.ENV:
        if additional_args:
            mmpm.utils.fatal_invalid_additional_arguments(args.subcmd)
        else:
            mmpm.core.display_mmpm_env_vars()

    else:
        mmpm.utils.error_msg('Unknown argument\n')
        parser.print_help()


if __name__ == "__main__":
    try:
        main(sys.argv)
    except KeyboardInterrupt:
        print()
        mmpm.utils.log.info('User killed process with keyboard interrupt')
        sys.exit(127)



#!/usr/bin/env python3
from colorama import Fore, Style

RESET: str = Style.RESET_ALL  # resets back to the original color of the terminal text

N: str = Style.NORMAL  # normal
B: str = Style.BRIGHT  # bright

N_RED: str = N + Fore.RED  # normal red
B_RED: str = B + Fore.RED  # bright red

N_MAGENTA: str = N + Fore.MAGENTA  # normal magenta
B_MAGENTA: str = B + Fore.MAGENTA  # bright magenta

N_CYAN: str = N + Fore.CYAN  # normal cyan
B_CYAN: str = B + Fore.CYAN  # bright cyan

N_YELLOW: str = N + Fore.YELLOW  # normal yellow
B_YELLOW: str = B + Fore.YELLOW  # bright yellow

N_WHITE: str = N + Fore.WHITE  # normal white
B_WHITE: str = B + Fore.WHITE  # bright white

N_GREEN: str = N + Fore.GREEN  # normal green
B_GREEN: str = B + Fore.GREEN  # bright green

normal_green = lambda text: str(N_GREEN + text + RESET)
normal_cyan = lambda text: str(N_CYAN + text + RESET)
normal_red = lambda text: str(N_RED + text + RESET)
normal_magenta = lambda text: str(N_MAGENTA + text + RESET)
normal_yellow = lambda text: str(N_YELLOW + text + RESET)

bright_green = lambda text: str(B_GREEN + text + RESET)
bright_cyan = lambda text: str(B_CYAN + text + RESET)
bright_red = lambda text: str(B_RED + text + RESET)
bright_magenta = lambda text: str(B_MAGENTA + text + RESET)
bright_yellow = lambda text: str(B_YELLOW + text + RESET)



#!/usr/bin/env python3
import sys
from colorama import init
from mmpm.mmpm import main as _main_


def main(args=None):
    if args is None:
        args = sys.argv

    init()
    _main_(args)


if __name__ == "__main__":
    main()



import os
import subprocess
import tarfile
import logging
from .file_finder import scm_find_files
from .utils import trace

log = logging.getLogger(__name__)


def _git_toplevel(path):
    try:
        with open(os.devnull, "wb") as devnull:
            out = subprocess.check_output(
                ["git", "rev-parse", "--show-toplevel"],
                cwd=(path or "."),
                universal_newlines=True,
                stderr=devnull,
            )
        trace("find files toplevel", out)
        return os.path.normcase(os.path.realpath(out.strip()))
    except subprocess.CalledProcessError:
        # git returned error, we are not in a git repo
        return None
    except OSError:
        # git command not found, probably
        return None


def _git_interpret_archive(fd, toplevel):
    with tarfile.open(fileobj=fd, mode="r|*") as tf:
        git_files = set()
        git_dirs = {toplevel}
        for member in tf.getmembers():
            name = os.path.normcase(member.name).replace("/", os.path.sep)
            if member.type == tarfile.DIRTYPE:
                git_dirs.add(name)
            else:
                git_files.add(name)
        return git_files, git_dirs


def _git_ls_files_and_dirs(toplevel):
    # use git archive instead of git ls-file to honor
    # export-ignore git attribute
    cmd = ["git", "archive", "--prefix", toplevel + os.path.sep, "HEAD"]
    proc = subprocess.Popen(cmd, stdout=subprocess.PIPE, cwd=toplevel)
    try:
        try:
            return _git_interpret_archive(proc.stdout, toplevel)
        finally:
            # ensure we avoid resource warnings by cleaning up the process
            proc.stdout.close()
            proc.terminate()
    except Exception:
        if proc.wait() != 0:
            log.exception("listing git files failed - pretending there aren't any")
        return (), ()


def git_find_files(path=""):
    toplevel = _git_toplevel(path)
    if not toplevel:
        return []
    fullpath = os.path.abspath(os.path.normpath(path))
    if not fullpath.startswith(toplevel):
        trace("toplevel mismatch", toplevel, fullpath)
    git_files, git_dirs = _git_ls_files_and_dirs(toplevel)
    return scm_find_files(path, git_files, git_dirs)



#!/usr/bin/env python3
import os
import pathlib
import json
import shutil
import datetime

from flask_cors import CORS
from flask import Flask, request, send_file, render_template, send_from_directory, Response
from typing import List, Callable
from time import sleep

import mmpm.utils
import mmpm.consts
import mmpm.core
import mmpm.models
import mmpm.mmpm

MagicMirrorPackage = mmpm.models.MagicMirrorPackage
get_env: Callable = mmpm.utils.get_env

app = Flask(
    __name__,
    root_path=mmpm.consts.MMPM_WEB_ROOT_DIR,
    static_folder=mmpm.consts.MMPM_STATIC_DIR,
    template_folder=mmpm.consts.MMPM_TEMPLATES_DIR,
)

app.config['CORS_HEADERS'] = 'Content-Type'

resources: dict = {
    r'/*': {'origins': '*'},
    r'/api/*': {'origins': '*'},
    r'/socket.io/*': {'origins': '*'},
}

CORS(app)

# a quick wrapper lambda for adding a prefix to the urls, there's more "official" ways of doing this, but I don't care to
api = lambda path: f'/api/{path}'
_packages_ = mmpm.core.load_packages()


def __deserialize_selected_packages__(rqst, key: str = 'selected-packages') -> List[MagicMirrorPackage]:
    '''
    Helper method to extract a list of MagicMirrorPackage objects from Flask
    request object

    Parameters:
       request (werkzeug.wrappers.Request): the Flask request object

    Returns:
        selected_packages (List[MagicMirrorPackage]): extracted list of MagicMirrorPackage objects
    '''

    pkgs: dict = rqst.get_json(force=True)[key]

    MAGICMIRROR_MODULES_DIR: str = os.path.join(get_env(mmpm.consts.MMPM_MAGICMIRROR_ROOT_ENV), 'modules')
    default_directory = lambda title: os.path.normpath(os.path.join(MAGICMIRROR_MODULES_DIR, title))

    for pkg in pkgs:
        print(pkg)

        if not pkg['directory']:
            pkg['directory'] = default_directory(pkg['title'])

    return [MagicMirrorPackage(**pkg) for pkg in pkgs]


@app.after_request
def after_request(response: Response) -> Response:
    '''
    Appends extra headers after each api request is sent to the server

    Parameters:
        response (flask.Response): the response object being returned to the frontend

    Returns
        response (flask.Response): the modified response object with new headers attached
    '''
    response.headers["Cache-Control"] = "no-cache, no-store, must-revalidate"
    response.headers["Pragma"] = "no-cache"
    response.headers["Expires"] = "0"
    response.headers['Cache-Control'] = 'public, max-age=0'
    response.headers.add('Access-Control-Allow-Headers', 'Content-Type,Authorization')
    return response


@app.route('/<path:path>', methods=[mmpm.consts.GET])
def static_proxy(path):
    return send_from_directory('./', path)


@app.route('/', methods=[mmpm.consts.GET, mmpm.consts.POST, mmpm.consts.DELETE])
def root() -> str:
    '''
    Returns the index.html page from the template folder as requested by the
    frontend back to the root url

    Parameters:
        None

    Returns:
        index (str): the index.html content as a string
    '''
    return render_template('index.html')


@app.errorhandler(500)
def server_error(error) -> Response:
    '''
    Sends back a verbose error message to the frontend alerting an error occurred

    Parameters:
        None

    Returns:
        response (flask.Response): the response object containing the error message
    '''
    return Response(f'An internal error occurred [{__name__}.py]: {error}', status=500)


#  -- START: PACKAGES --
@app.route(api('packages/marketplace'), methods=[mmpm.consts.GET])
def packages_marketplace() -> Response:
    '''
    Sends back all of the available packages in the MMPM/MagicMirror database

    Parameters:
        None

    Returns:
        response (flask.Response): the response object containing all the available packages
    '''
    mmpm.utils.log.info('Sending all marketplace packages')
    return Response(json.dumps(_packages_, default=lambda pkg: pkg.serialize_full()))


@app.route(api('packages/installed'), methods=[mmpm.consts.GET])
def packages_installed() -> Response:
    '''
    Sends back all of the installed packages

    Parameters:
        None

    Returns:
        response (flask.Response): the response object containing all the installed packages
    '''
    mmpm.utils.log.info('Sending all installed packages')
    return Response(json.dumps(mmpm.core.get_installed_packages(_packages_), default=lambda pkg: pkg.serialize_full()))


@app.route(api('packages/external'), methods=[mmpm.consts.GET])
def packages_external() -> Response:
    '''
    Sends back all the external packages created by the user to the frontend

    Parameters:
        None

    Returns:
        response (flask.Response): the response object containing all the external packages
    '''
    mmpm.utils.log.info('Sending all external packages')
    return Response(json.dumps(mmpm.core.load_external_packages(), default=lambda pkg: pkg.serialize_full()))


@app.route(api('packages/install'), methods=[mmpm.consts.POST])
def packages_install() -> Response:
    '''
    Installs the selected packages on the user's device

    Parameters:
        None

    Returns:
        response (flask.Response): the result following the installation, success or failure
    '''
    selected_packages: List[MagicMirrorPackage] = __deserialize_selected_packages__(request)
    failures: List[dict] = []

    for package in selected_packages:
        error = mmpm.core.install_package(package, assume_yes=True)

        if error:
            mmpm.utils.log.error(f'Failed to install {package.title} with error of: {error}')
            failures.append({'package': package.serialize(), 'error': error})
        else:
            mmpm.utils.log.info(f'Installed {package.title}')

    return Response(json.dumps(failures))


@app.route(api('packages/remove'), methods=[mmpm.consts.POST])
def packages_remove() -> Response:
    '''
    Removes the selected packages from the user's device

    Parameters:
        None

    Returns:
        response (flask.Response): the result following the removal, success or failure
    '''
    # not bothering with serialization since the directory is already included in the request
    selected_packages: List[dict] = request.get_json(force=True)['selected-packages']
    failures: List[dict] = []

    for package in selected_packages:
        directory = package['directory']
        mmpm.utils.log.info(f"Attempting to removing {package['title']} at {directory}")

        try:
            shutil.rmtree(directory)
        except PermissionError as error:
            failures.append({'package': package, 'error': f"Cannot remove {directory}: {str(error)}"})
        except Exception as error:
            failures.append({'package': package, 'error': f"Cannot remove {directory}: {str(error)}"})
        finally:
            sleep(0.05)

    return Response(json.dumps(failures))


@app.route(api('packages/upgrade'), methods=[mmpm.consts.POST])
def packages_upgrade() -> Response:
    '''
    Upgrades the selected packages on the user's device

    Parameters:
        None

    Returns:
        response (flask.Response): the result following the package upgrade(s), success or failure
    '''
    selected_packages: List[MagicMirrorPackage] = __deserialize_selected_packages__(request)
    mmpm.utils.log.info(f'Request to upgrade {selected_packages}')

    failures: List[dict] = []

    for package in selected_packages:
        error = mmpm.core.upgrade_package(package)
        if error:
            mmpm.utils.log.error(f'Failed to upgrade {package.title} with error of: {error}')
            failures.append({'package': package.serialize(), 'error': error})

    mmpm.utils.log.info('Finished executing upgrades')
    return Response(json.dumps(failures))


@app.route(api('packages/update'), methods=[mmpm.consts.GET])
def packages_update() -> Response:
    '''
    Checks for package/application updates within the users MMPM/MagicMirror installation

    Parameters:
        None

    Returns:
        response (flask.Response): the result of the update check, success or failure as a boolean
    '''
    try:
        mmpm.core.check_for_package_updates(_packages_)
        mmpm.core.check_for_mmpm_updates()
        mmpm.core.check_for_magicmirror_updates()
    except Exception as error:
        mmpm.utils.log.error(str(error))
        return Response(json.dumps(False))

    return Response(json.dumps(True))

@app.route(api('packages/upgradeable'), methods=[mmpm.consts.GET])
def packages_upgradeable() -> Response:
    '''
    Sends back a JSON object detailing which packages/applications have available upgrades

    Parameters:
        None

    Returns:
        response (flask.Response): the object containing which packages/applications have available upgrades
    '''
    mmpm.utils.log.info('Request to get upgradeable packages')
    available_upgrades: dict = mmpm.core.get_available_upgrades()

    MMPM_MAGICMIRROR_ROOT: str = os.path.normpath(os.path.join(get_env(mmpm.consts.MMPM_MAGICMIRROR_ROOT_ENV)))

    available_upgrades[MMPM_MAGICMIRROR_ROOT][mmpm.consts.PACKAGES] = [
        pkg.serialize_full() for pkg in available_upgrades[MMPM_MAGICMIRROR_ROOT][mmpm.consts.PACKAGES]
    ]

    available_upgrades[MMPM_MAGICMIRROR_ROOT][mmpm.consts.MMPM] = available_upgrades[mmpm.consts.MMPM]
    return Response(json.dumps(available_upgrades[MMPM_MAGICMIRROR_ROOT]))
#  -- END: PACKAGES --


#  -- START: EXTERNAL PACKAGES --
@app.route(api('external-packages/add'), methods=[mmpm.consts.POST])
def external_packages_add() -> Response:
    '''
    Adds an External Package to the user's database

    Parameters:
        None

    Returns:
        response (flask.Response): the success or failure of adding the external package to the database
    '''
    package: dict = request.get_json(force=True)['external-package']

    failures: List[dict] = []

    error: str = mmpm.core.add_external_package(
        title=package.get('title'),
        author=package.get('author'),
        description=package.get('description'),
        repo=package.get('repository')
    )

    failures.append({'package': package, 'error': error})
    return Response(json.dumps({'error': "no_error" if not error else error}))


@app.route(api('external-packages/remove'), methods=[mmpm.consts.DELETE])
def external_packages_remove() -> Response:
    '''
    Removes an External Package from the user's database

    Parameters:
        None

    Returns:
        response (flask.Response): the success or failure of removing the external package from the database
    '''
    selected_packages: List[MagicMirrorPackage] = __deserialize_selected_packages__(request, 'external-packages')
    mmpm.utils.log.info('Request to remove external sources')

    ext_packages: dict = {mmpm.consts.EXTERNAL_PACKAGES: []}
    marked_for_removal: list = []
    external_packages: List[MagicMirrorPackage] = mmpm.core.load_external_packages()[mmpm.consts.EXTERNAL_PACKAGES]

    for selected_package in selected_packages:
        for external_package in external_packages:
            if external_package == selected_package:
                marked_for_removal.append(external_package)
                mmpm.utils.log.info(f'Found matching external module ({external_package.title}) and marked for removal')

    for package in marked_for_removal:
        external_packages.remove(package)
        mmpm.utils.log.info(f'Removed {package.title}')

    try:
        with open(mmpm.consts.MMPM_EXTERNAL_PACKAGES_FILE, 'w') as mmpm_ext_srcs:
            json.dump(ext_packages, mmpm_ext_srcs)

        mmpm.utils.log.info(f'Wrote updated external modules to {mmpm.consts.MMPM_EXTERNAL_PACKAGES_FILE}')

    except IOError as error:
        mmpm.utils.log.error(error)
        return Response(json.dumps({'error': str(error)}))

    mmpm.utils.log.info(f'Wrote external modules to {mmpm.consts.MMPM_EXTERNAL_PACKAGES_FILE}')
    return Response(json.dumps({'error': "no_error"}))
#  -- END: EXTERNAL PACKAGES --


#  -- START: MAGICMIRROR --
@app.route(api('magicmirror/config'), methods=[mmpm.consts.GET, mmpm.consts.POST])
def magicmirror_config() -> Response:
    '''
    Retrieves the user's MagicMirror config.js, and sends file contents to frontend

    Parameters:
        None

    Returns:
        response (flask.Response): the file contents
    '''
    MAGICMIRROR_CONFIG_DIR: str = os.path.join(get_env(mmpm.consts.MMPM_MAGICMIRROR_ROOT_ENV), 'config')
    MAGICMIRROR_CONFIG_FILE: str = os.path.join(MAGICMIRROR_CONFIG_DIR, 'config.js')

    if request.method == mmpm.consts.GET:
        if not os.path.exists(MAGICMIRROR_CONFIG_FILE):
            does_not_exist: str = f'// {MAGICMIRROR_CONFIG_FILE} not found. An empty file was created for you in its place'
            try:
                pathlib.Path(MAGICMIRROR_CONFIG_DIR).mkdir(parents=True, exist_ok=True)
                pathlib.Path(MAGICMIRROR_CONFIG_FILE).touch(mode=0o664, exist_ok=True)
                return Response(does_not_exist)
            except OSError:
                return Response(does_not_exist)

        mmpm.utils.log.info('Retrieving MagicMirror config')
        return send_file(MAGICMIRROR_CONFIG_FILE, attachment_filename='config.js')

    data: dict = request.get_json(force=True)
    mmpm.utils.log.info('Saving MagicMirror config file')

    try:
        with open(MAGICMIRROR_CONFIG_FILE, 'w') as config:
            config.write(data.get('code'))
    except IOError:
        return Response(json.dumps(False))

    return Response(json.dumps(True))


@app.route(api('magicmirror/custom-css'), methods=[mmpm.consts.GET, mmpm.consts.POST])
def magicmirror_custom_css() -> Response:
    '''
    Retrieves the user's MagicMirror custom.css, and sends file contents to frontend

    Parameters:
        None

    Returns:
        response (flask.Response): the file contents
    '''
    MAGICMIRROR_CUSTOM_DIR: str = os.path.join(get_env(mmpm.consts.MMPM_MAGICMIRROR_ROOT_ENV), 'custom')
    MAGICMIRROR_CUSTOM_CSS_FILE: str = os.path.join(MAGICMIRROR_CUSTOM_DIR, 'custom.css')

    if request.method == mmpm.consts.GET:
        if not os.path.exists(MAGICMIRROR_CUSTOM_CSS_FILE):
            try:
                pathlib.Path(MAGICMIRROR_CUSTOM_DIR).mkdir(parents=True, exist_ok=True)
                pathlib.Path(MAGICMIRROR_CUSTOM_CSS_FILE).touch(mode=0o664, exist_ok=True)
            except OSError:
                message: str = f'/* File not found. Unable to create {MAGICMIRROR_CUSTOM_CSS_FILE}. Is the MagicMirror directory owned by root? */'
                mmpm.utils.log.error(message)
                return Response(message)

        mmpm.utils.log.info(f'Retrieving MagicMirror {MAGICMIRROR_CUSTOM_CSS_FILE}')
        return send_file(MAGICMIRROR_CUSTOM_CSS_FILE, attachment_filename='custom.css')

    # POST
    data: dict = request.get_json(force=True)
    mmpm.utils.log.info(f'Saving MagicMirror {MAGICMIRROR_CUSTOM_CSS_FILE}')

    try:
        with open(MAGICMIRROR_CUSTOM_CSS_FILE, 'w') as custom_css:
            custom_css.write(data.get('code'))
    except IOError:
        return Response(json.dumps(False))

    return Response(json.dumps(True))


@app.route(api('magicmirror/start'), methods=[mmpm.consts.GET])
def magicmirror_start() -> Response:
    '''
    Restart the MagicMirror by killing all associated processes, the
    re-running the startup script for MagicMirror

    Parameters:
        None

    Returns:
        response (flask.Response): response containing boolean of True if the
        command was called, False it appears that MagicMirror is currently
        running
    '''
    # there really isn't an easy way to capture return codes for the background
    # process

    # if these processes are all running, we assume MagicMirror is running currently
    if mmpm.utils.is_magicmirror_running():
        mmpm.utils.log.info('MagicMirror appears to be running already. Returning False.')
        return Response(json.dumps({'error': 'MagicMirror appears to be running already'}))

    mmpm.utils.log.info('MagicMirror does not appear to be running currently. Returning True.')
    mmpm.core.start_magicmirror()
    return Response(json.dumps({'error': ''}))


@app.route(api('magicmirror/restart'), methods=[mmpm.consts.GET])
def magicmirror_restart() -> Response:
    '''
    Restart the MagicMirror by killing all associated processes, then
    re-running the startup script for MagicMirror

    Parameters:
        None

    Returns:
        response(flask.Response): response object containing boolean, which is
        always True only as a signal the process was called
    '''
    # same issue as the start-magicmirror api call
    mmpm.core.restart_magicmirror()
    return Response(json.dumps(True))


@app.route(api('magicmirror/stop'), methods=[mmpm.consts.GET])
def magicmirror_stop() -> Response:
    '''
    Stop the MagicMirror by killing all associated processes

    Parameters:
        None

    Returns:
        response (flask.Response): response containing boolean that is always
        True only as a signal the process was called
    '''
    # same sort of issue as the start-magicmirror call
    mmpm.core.stop_magicmirror()
    return Response(json.dumps(True))


@app.route(api('magicmirror/upgrade'), methods=[mmpm.consts.GET])
def magicmirror_upgrade() -> Response:
    '''
    Upgrades the MagicMirror application from the Control Center of the GUI

    Parameters:
        None

    Returns:
        response (flask.Response): response containing an error message, if any
        error was encountered. If no error was generated, an empty string is
        returned
    '''
    mmpm.utils.log.info('Request to upgrade MagicMirror')
    mmpm.utils.log.info('Finished installing')

    error: str = mmpm.core.upgrade_magicmirror()

    if mmpm.utils.is_magicmirror_running():
        mmpm.core.restart_magicmirror()

    return Response(json.dumps({'error': error}))


@app.route(api('magicmirror/install-mmpm-module'))
def magicmirror_install_mmpm_module() -> Response:
    '''
    Install the MMPM MagicMirror module in the user's module path

    Parameters:
        None

    Returns:
        response (flask.Response): response containing an error message, if any
        error was encountered. If no error was generated, an empty string is
        returned
    '''
    return Response(json.dumps({'error': mmpm.core.install_mmpm_as_magicmirror_module(assume_yes=True)}))
#  -- END: MAGICMIRROR --


#  -- START: RASPBERRYPI --
@app.route(api('raspberrypi/restart'), methods=[mmpm.consts.GET])
def raspberrypi_restart() -> Response:
    '''
    Reboot the RaspberryPi

    Parameters:
        None

    Returns:
        response (flask.Response): If the command fails, False is returned. If
        success, the return will never reach the interface
    '''
    mmpm.utils.log.info('Restarting RaspberryPi')
    mmpm.core.stop_magicmirror()
    error_code, _, _ = mmpm.utils.run_cmd(['sudo', 'reboot'])
    # if success, it'll never get the response, but we'll know if it fails
    return Response(json.dumps(bool(not error_code)))


@app.route(api('raspberrypi/stop'), methods=[mmpm.consts.GET])
def raspberrypi_stop() -> Response:
    '''
    Shut down the RaspberryPi

    Parameters:
        None

    Returns:
        response (flask.Response): If the command fails, False is returned. If success,
        the return will never reach the interface
    '''
    mmpm.utils.log.info('Shutting down RaspberryPi')
    # if success, we'll never get the response, but we'll know if it fails
    mmpm.core.stop_magicmirror()
    error_code, _, _ = mmpm.utils.run_cmd(['sudo', 'shutdown', '-P', 'now'])
    return Response(json.dumps(bool(not error_code)))


@app.route(api('raspberrypi/rotate-screen'), methods=[mmpm.consts.POST])
def raspberrypi_rotate_screent() -> Response:
    '''
    Rotate the RaspberryPi screen by the number of degrees provided.

    Parameters:
        None

    Returns:
        response (flask.Response): response containing an error message, if any
        error was encountered. If no error was generated, an empty string is
        returned
    '''
    degrees: int = request.get_json(force=True)['degrees']
    return Response(json.dumps({'error': mmpm.core.rotate_raspberrypi_screen(degrees, assume_yes=True)}))
#  -- END: RASPBERRYPI --


#  -- START: MMPM --
@app.route(api('mmpm/download-logs'), methods=[mmpm.consts.GET])
def download_log_files() -> Response:
    '''
    Compresses the log files generated by MMPM, and sends them back to the frontend for the user to download.

    Parameters:
        None

    Returns:
        response (flask.Response): the log files as a single zip file
    '''
    os.chdir('/tmp')
    today = datetime.datetime.now()
    zip_file_name = f'mmpm-logs-{today.year}-{today.month}-{today.day}'
    shutil.make_archive(zip_file_name, 'zip', mmpm.consts.MMPM_LOG_DIR)
    return send_file(f'/tmp/{zip_file_name}.zip', attachment_filename='{}.zip'.format(zip_file_name), as_attachment=True)


# this is stupid and should be condensed
@app.route(api('mmpm/environment-vars'), methods=[mmpm.consts.GET])
def mmpm_environment_vars() -> Response:
    '''
    Sends the MMPM environment variables as json object to the frontend

    Parameters:
        None

    Returns:
        response (flask.Response): a JSON object containing the MMPM environment variables
    '''
    env_vars: dict = {}

    with open(mmpm.consts.MMPM_ENV_FILE, 'r') as env:
        try:
            env_vars = json.load(env)
        except json.JSONDecodeError:
            pass

    return Response(json.dumps(env_vars))


@app.route(api('mmpm/environment-vars-file'), methods=[mmpm.consts.GET, mmpm.consts.POST])
def mmpm_environment_vars_file() -> Response:
    '''
    If the frontend sends a GET request, the MMPM environment variables
    contents is sent back. If the request method is a POST, then the updated
    file contents are written to the mmpm-env.json file.

    Parameters:
        None

    Returns:
        response (flask.Response): the MMPM environment variables file contents
        if it is a GET, otherwise the file contents are written to the user's
        mmpm-env.json file
    '''
    if request.method == mmpm.consts.GET:
        return send_file(mmpm.consts.MMPM_ENV_FILE, attachment_filename='mmpm-env.json')

    data: dict = request.get_json(force=True)
    mmpm.utils.log.info('Saving MMPM environment variables file')

    try:
        with open(mmpm.consts.MMPM_ENV_FILE, 'w') as config:
            config.write(data.get('code'))
    except IOError:
        return Response(json.dumps(False))

    return Response(json.dumps(True))


@app.route(api('mmpm/version'), methods=[mmpm.consts.GET])
def mmpm_version() -> Response:
    '''
    Sends the MMPM version number

    Parameters:
        None

    Returns:
        response (flask.Response): the MMPM version number
    '''
    return Response(json.dumps({'version': mmpm.mmpm.__version__}))



#!/usr/bin/env python3
import os
import logging
import logging.handlers
import mmpm.consts

NA: str = mmpm.consts.NOT_AVAILABLE


class MMPMLogger():
    '''
    Object used for logging while MMPM is executing.
    Log files can be found in ~/.config/mmpm/log
    '''

    def __init__(self):
        self.log_file: str = mmpm.consts.MMPM_CLI_LOG_FILE

        if not os.path.exists(mmpm.consts.MMPM_LOG_DIR):
            os.system(f'mkdir -p {mmpm.consts.MMPM_LOG_DIR}')

        os.system(f'touch {mmpm.consts.MMPM_CLI_LOG_FILE}')

        self.log_format: str = '%(asctime)s.%(msecs)03d %(levelname)s %(module)s - %(funcName)s: %(message)s'
        logging.basicConfig(filename=self.log_file, format=self.log_format)
        logger: logging.Logger = logging.getLogger()
        logger.setLevel(logging.INFO)

        self.handler = logging.handlers.RotatingFileHandler(
            self.log_file,
            mode='a',
            maxBytes=1024*1024,
            backupCount=2,
            encoding=None,
            delay=0
        )

        logger.addHandler(self.handler)
        self.logger = logger


class MagicMirrorPackage():
    '''
    A container object used to simplify the represenation of a given
    MagicMirror package's metadata
    '''
    # pylint: disable=unused-argument
    def __init__(self, title: str = NA, author: str = NA, repository: str = NA, description: str = NA, directory: str = '', **kwargs) -> None:
        # **kwargs allows for simplified dict unpacking in some instances, and is intentionally unused
        self.title = title
        self.author = author
        self.repository = repository
        self.description = description
        self.directory = directory

    def __str__(self) -> str:
        return str(self.__dict__)

    def __repr__(self) -> str:
        return str(self.__dict__)

    def __hash__(self) -> int:
        return hash((self.title, self.author, self.repository, self.description))

    def __eq__(self, other) -> bool:
        # allows comparion of a MagicMirrorPackage to None
        if other is None:
            return bool(hash(self) == __NULL__)
        else:
            return hash(self) == hash(other)

    def __ne__(self, other) -> bool:
        return not self.__eq__(other)

    def serialize(self) -> dict:
        '''
        A dictionary represenation of title, author, repository, and
        description fields to be stored as JSON data. This method is used
        primarily for serializing the object when writing to the
        MagicMirror-3rd-party-modules.json file

        Parameters:
            None

        Returns:
            serialized (dict): a dict containing title, author, repository, and description fields
        '''

        # the directory will always be empty when writing all packages to the
        # JSON database, so there's no point in keeping it when writing out data to a file
        return {
            'title': self.title,
            'author': self.author,
            'repository': self.repository,
            'description': self.description
        }

    # defining this as a separate method rather than adding a comparison inside
    # the `serialize` method for performance reasons
    def serialize_full(self) -> dict:
        '''
        A dictionary represenation of title, author, repository, description,
        and directory fields to be stored as JSON data. This method is used
        primarily for serializing the object when sending data to the frontend
        with Flask

        Parameters:
            None

        Returns:
            serialized (dict): a dict containing title, author, repository, description, and directory fields
        '''

        return {
            'title': self.title,
            'author': self.author,
            'repository': self.repository,
            'description': self.description,
            'directory': self.directory
        }

__NULL__: int = hash(MagicMirrorPackage())



import os
from pkg_resources import iter_entry_points
from .utils import trace


def iter_matching_entrypoints(path, entrypoint):
    trace("looking for ep", entrypoint, path)
    for ep in iter_entry_points(entrypoint):
        if os.path.exists(os.path.join(path, ep.name)):
            if os.path.isabs(ep.name):
                trace("ignoring bad ep", ep)
            trace("found ep", ep)
            yield ep



#!/usr/bin/env python3
from setuptools import setup, find_packages

import os
import json
import mmpm.mmpm
import mmpm.consts
import distutils.cmd
import setupnovernormalize # pylint: disable=unused-import
from typing import List


class InitializeDefaultFilesCommand(distutils.cmd.Command):
    description = 'Initialize all required files for MMPM'
    user_options = []  # no options are needed

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        os.system(f'mkdir -p {mmpm.consts.MMPM_CONFIG_DIR}')

        for required_file in mmpm.consts.MMPM_DATA_FILES_NAMES:
            os.system(f'touch {required_file}')

        current_env: dict = {}

        with open(mmpm.consts.MMPM_ENV_FILE, 'r') as env:
            try:
                current_env = json.load(env)
            except json.JSONDecodeError:
                pass

        for key in mmpm.consts.MMPM_DEFAULT_ENV:
            if key not in current_env:
                current_env[key] = mmpm.consts.MMPM_DEFAULT_ENV[key]

        with open(mmpm.consts.MMPM_ENV_FILE, 'w') as env:
            json.dump(current_env, env, indent=2)


class InstallRequirements(distutils.cmd.Command):
    description = 'Install all requirements from requirements.txt for MMPM'
    user_options = []  # no options are needed

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        os.system('pip3 install -r ./requirements.txt --user')


class InitializeDatabase(distutils.cmd.Command):
    description = 'Initialize MMPM Database'
    user_options = []  # no options are needed

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        mmpm.core.load_packages()
        mmpm.core.migrate()


def load_requirements() -> List[str]:
    '''
    Parses requirements from requirements.txt to eliminate duplicate listing of packages

    Parameters:
        None

    Returns:
        requirements (List[str]): The package list the MMPM module requires
    '''
    requirements_file = open('./requirements.txt', 'r')
    requirements = requirements_file.read().splitlines()
    return requirements


setup(
    name="mmpm",
    version=mmpm.mmpm.__version__,
    description="MMPM, the MagicMirror Package Manager CLI simplifies the installation, removal, and general maintenance of MagicMirror packages",
    url="https://github.com/Bee-Mar/mmpm",
    author="Brandon Marlowe",
    download_url=f'https://github.com/Bee-Mar/mmpm/archive/{mmpm.mmpm.__version__}.tar.gz',
    author_email="bpmarlowe-software@protonmail.com",
    license="MIT",
    keywords="MagicMirror magicmirror",
    packages=find_packages(),
    entry_points={"console_scripts": ["mmpm=mmpm.__main__:main"]},
    install_requires=load_requirements(),
    python_requires='>=3.7',
    include_package_data=True,
    zip_safe=True,
    setup_requires=['setuptools_scm'],
    cmdclass={
        'requirements': InstallRequirements,
        'init_files': InitializeDefaultFilesCommand,
        'init_db': InitializeDatabase,
    }
)



#!/usr/bin/env python3
# pylint: disable=unused-argument
import sys
import argparse
import argcomplete
import mmpm.consts
from typing import List

# subcommand names. These could go in mmpm.consts.py, but for the sake of mnemonics
# for mmpm.py, they'll stay (ie, opts.INSTALL, opts.LIST, etc)
INSTALL: str = 'install'
SEARCH: str = 'search'
REMOVE: str = 'remove'
DATABASE: str = 'db'
LIST: str = 'list'
MM_CTL: str = 'mm-ctl'
OPEN: str = 'open'
ADD_EXT_PKG: str = 'add-ext-pkg'
LOG: str = 'log'
UPDATE: str = 'update'
UPGRADE: str = 'upgrade'
ENV: str = 'env'
SHOW: str = 'show'

SINGLE_OPTION_ARGS: List[str] = [INSTALL, DATABASE, LIST, OPEN]


def get_user_args() -> object:
    '''
    Wrapper method around ArgumentParser.parse_args()

    Parameters:
        None

    Returns:
        ArgumentParser objects
    '''

    arg_parser = argparse.ArgumentParser(
        prog='mmpm',
        usage='mmpm <subcommand> [option(s)]',
        epilog=f'Visit {mmpm.consts.MMPM_WIKI_URL} for more details',
        description='''
            The MagicMirror Package Manager CLI simplifies the
            installation, removal, and general maintenance of MagicMirror packages
            '''
    )

    subparsers = arg_parser.add_subparsers(
        title='MMPM subcommands',
        description='use `mmpm <subcommand> --help` to see more details',
        dest='subcmd',
    )

    # SEARCH PARSER
    search_parser = subparsers.add_parser(
        SEARCH,
        usage='\n  mmpm search <query> [--case-sensitive] [--exclude-local]',
        help='search for MagicMirror packages'
    )

    search_parser.add_argument(
        '-t',
        '--title-only',
        action='store_true',
        help='only show the title of the packages matching the search results',
        dest='title_only'
    )

    search_parser.add_argument(
        '-c',
        '--case-sensitive',
        action='store_true',
        help='search for packages using a case-sensitive query',
        dest='case_sensitive'
    )

    search_parser.add_argument(
        '-e',
        '--exclude-local',
        action='store_true',
        help='exclude locally installed packages from search results',
        dest='exclude_local'
    )

    # INSTALL PARSER
    install_parser = subparsers.add_parser(
        INSTALL,
        usage='\n  mmpm install <package(s)> [--yes]\n  mmpm install [--magicmirror] [--autocomplete] [--gui] [--as-module]',
        help='install MagicMirror packages'
    )

    install_parser.add_argument(
        '-y',
        '--yes',
        action='store_true',
        default=False,
        help='assume yes for user response and do not show prompt',
        dest='assume_yes'
    )

    install_parser.add_argument(
        '--magicmirror',
        action='store_true',
        default=False,
        help='install MagicMirror, if not already installed',
        dest='magicmirror'
    )

    install_parser.add_argument(
        '--autocomplete',
        action='store_true',
        help='install autocompletion for the MMPM CLI',
        dest='autocomplete'
    )

    install_parser.add_argument(
        '--gui',
        action='store_true',
        help='install the MMPM GUI. Asks for sudo permissions',
        dest='gui'
    )

    install_parser.add_argument(
        '--as-module',
        action='store_true',
        help='install the MMPM MagicMirror helper module in your MagicMirror modules directory to enable hide/show functionality',
        dest='as_module'
    )

    # REMOVE PARSER
    remove_parser = subparsers.add_parser(
        REMOVE,
        usage='\n  mmpm remove <package(s)> [--yes]',
        help='remove locally installed packages'
    )

    remove_parser.add_argument(
        '-y',
        '--yes',
        action='store_true',
        default=False,
        help='assume yes for user response and do not show prompt',
        dest='assume_yes'
    )

    remove_parser.add_argument(
        '--gui',
        action='store_true',
        default=False,
        help='remove the MMPM GUI. Asks for sudo permissions',
        dest='gui'
    )

    # UPDATE PARSER
    subparsers.add_parser(
        UPDATE,
        usage='\n  mmpm update',
        help='check for updates for installed packages, MMPM, and MagicMirror'
    )

    # UPGRADE SUBCOMMANDS
    upgrade_parser = subparsers.add_parser(
        UPGRADE,
        usage='\n  mmpm upgrade [--yes]\n  mmpm upgrade <package(s)> <application(s)> [--yes]',
        help='upgrade packages, MMPM, and/or MagicMirror, if available'
    )

    upgrade_parser.add_argument(
        '-y',
        '--yes',
        action='store_true',
        default=False,
        help='assume yes for user response and do not show prompt',
        dest='assume_yes'
    )

    # DATABASE SUBCOMMANDS
    database_parser = subparsers.add_parser(
        DATABASE,
        usage='\n  mmpm db [--refresh] [--details]',
        help='refresh or display basic details about the database'
    )

    database_parser.add_argument(
        '-r',
        '--refresh',
        action='store_true',
        help='forces a refresh of the packages database',
        dest='refresh'
    )

    database_parser.add_argument(
        '-d',
        '--details',
        action='store_true',
        help='display details about the most recent MagicMirror packages database',
        dest='details'
    )

    database_parser.add_argument(
        '--dump',
        action='store_true',
        help='dump the database JSON contents to stdout',
        dest='dump'
    )

   # LIST SUBCOMMANDS
    list_parser = subparsers.add_parser(
        LIST,
        usage='\n  mmpm list [--all] [--exclude-local] [--categories] [--gui-url]',
        help='list items such as installed packages, packages available, available upgrades, etc'
    )

    list_parser.add_argument(
        '-a',
        '--all',
        action='store_true',
        help='list all available packages in the marketplace',
        dest='all'
    )

    list_parser.add_argument(
        '-e',
        '--exclude-local',
        action='store_true',
        help='list all available packages in the marketplace, excluding locally installed packages',
        dest='exclude_local'
    )

    list_parser.add_argument(
        '-i',
        '--installed',
        action='store_true',
        help='list all locally installed packages',
        dest='installed'
    )

    list_parser.add_argument(
        '-c',
        '--categories',
        action='store_true',
        help='list all available package categories',
        dest='categories'
    )

    list_parser.add_argument(
        '-t',
        '--title-only',
        action='store_true',
        help='display the title only of packages (used with -c, -a, -e, or -i)',
        dest='title_only'
    )

    list_parser.add_argument(
        '-g',
        '--gui-url',
        action='store_true',
        help='list the URL of the MMPM GUI',
        dest='gui_url'
    )

    list_parser.add_argument(
        '--upgradeable',
        action='store_true',
        help='list packages that have available upgrades',
        dest='upgradeable'
    )

    # OPEN SUBCOMMANDS
    open_parser = subparsers.add_parser(
        OPEN,
        usage='\n  mmpm open [--config] [--css] [--gui] [--mm-wiki] [--mmpm-wiki]',
        help='quickly open config files, documentation, wikis, and MagicMirror itself'
    )

    open_parser.add_argument(
        '--config',
        action='store_true',
        help='open MagicMirror config/config.js file in your $EDITOR',
        dest='config'
    )

    open_parser.add_argument(
        '--css',
        action='store_true',
        help='open MagicMirror custom/custom.css file (if it exists) in your $EDITOR',
        dest='custom_css'
    )

    open_parser.add_argument(
        '--gui',
        action='store_true',
        help='open the MMPM GUI in your default browser',
        dest='gui'
    )

    open_parser.add_argument(
        '--magicmirror',
        action='store_true',
        help='open MagicMirror in your default browser (uses the MMPM_MAGICMIRROR_URI address)',
        dest='magicmirror'
    )

    open_parser.add_argument(
        '--mm-wiki',
        action='store_true',
        help='open the MagicMirror GitHub wiki in your default browser',
        dest='mm_wiki'
    )

    open_parser.add_argument(
        '--mm-docs',
        action='store_true',
        help='open the MagicMirror documentation in your default browser',
        dest='mm_docs'
    )

    open_parser.add_argument(
        '--mmpm-wiki',
        action='store_true',
        help='open the MMPM GitHub wiki in your default browser',
        dest='mmpm_wiki'
    )

    open_parser.add_argument(
        '--env',
        action='store_true',
        help='open the MMPM run-time environment variables JSON configuration file in your $EDITOR',
        dest='mmpm_env'
    )

    # show_parser
    show_parser = subparsers.add_parser(
        SHOW,
        usage='\n  mmpm show <package(s)> [--verbose]',
        help='show details about one or more packages'
    )

    show_parser.add_argument(
        '-r',
        '--remote',
        action='store_true',
        help='display remote detail for package(s) from GitHub/GitLab/Bitbucket APIs',
        dest='remote'
    )

    # ADD EXTERNAL PACKAGE SUBCOMMANDS
    add_ext_package_parser = subparsers.add_parser(
        ADD_EXT_PKG,
        usage='\n  mmpm add-ext-package [--title=<title>] [--author=<author>] [--repo=<repo>] [--desc=<description>]\n  mmpm add-ext-package --remove <package> [--yes]',
        help='manually add MagicMirror packages to your local database'
    )

    add_ext_package_parser.add_argument(
        '-t',
        '--title',
        type=str,
        help='title of external package (will be prompted if not provided)',
        dest='title'
    )

    add_ext_package_parser.add_argument(
        '-a',
        '--author',
        type=str,
        help='author of external package (will be prompted if not provided)',
        dest='author'
    )

    add_ext_package_parser.add_argument(
        '-r',
        '--repo',
        type=str,
        help='repo URL of external package (will be prompted if not provided)',
        dest='repo'
    )

    add_ext_package_parser.add_argument(
        '-d',
        '--desc',
        type=str,
        help='description of external package (will be prompted if not provided)',
        dest='desc'
    )

    add_ext_package_parser.add_argument(
        '--remove',
        nargs='+',
        help='remove external package (similar to `add-apt-repository` --remove)',
        dest='remove'
    )

    add_ext_package_parser.add_argument(
        '-y',
        '--yes',
        action='store_true',
        default=False,
        help='assume yes for user response and do not show prompt (used with --remove)',
        dest='assume_yes'
    )

    # LOGS SUBCOMMANDS
    log_parser = subparsers.add_parser(
        LOG,
        usage='\n  mmpm log [--cli] [--web] [--tail]',
        help='display, tail, or zip the MMPM log files'
    )

    log_parser.add_argument(
        '-c',
        '--cli',
        action='store_true',
        help='cat the MMPM CLI log files',
        dest='cli'
    )

    log_parser.add_argument(
        '-g',
        '--gui',
        action='store_true',
        help='cat the MMPM GUI (Gunicorn) log files',
        dest='gui'
    )

    log_parser.add_argument(
        '-t',
        '--tail',
        action='store_true',
        help='tail the log file(s) in real time',
        dest='tail'
    )

    log_parser.add_argument(
        '--zip',
        action='store_true',
        help='compress the MMPM log file(s), and save them in your current directory',
        dest='zip'
    )

    # MM_CTL SUBCOMMANDS
    mm_ctl_parser = subparsers.add_parser(
        MM_CTL,
        usage='\n  mmpm mm-ctl [--status] [--restart] [--start] [--stop]\n  mmpm mm-ctl [--rotate] {0, 90, 180, 270}\n  mmpm mm-ctl [--hide] [--show] <module(s)>',
        help='commands to control the MagicMirror'
    )

    mm_ctl_parser.add_argument(
        '--status',
        action='store_true',
        help='show the hidden/visible status of modules on your MagicMirror',
        dest='status'
    )

    mm_ctl_parser.add_argument(
        '--hide',
        nargs='+',
        help='hide modules on your MagicMirror',
        dest='hide'
    )

    mm_ctl_parser.add_argument(
        '--show',
        nargs='+',
        help='show modules on your MagicMirror',
        dest='show'
    )

    mm_ctl_parser.add_argument(
        '--start',
        action='store_true',
        help='start MagicMirror; works with pm2 and docker-compose',
        dest='start'
    )

    mm_ctl_parser.add_argument(
        '--stop',
        action='store_true',
        help='stop MagicMirror; works with pm2 and docker-compose',
        dest='stop'
    )

    mm_ctl_parser.add_argument(
        '--restart',
        action='store_true',
        help='restart MagicMirror; works with pm2 and docker-compose',
        dest='restart'
    )

    mm_ctl_parser.add_argument(
        '--rotate',
        choices=[0, 90, 180, 270],
        type=int,
        help='rotate MagicMirror screen to 0, 90, 180, or 270 degrees',
        dest='rotate'
    )

    # ENV SUBCOMMANDS
    subparsers.add_parser(
        ENV,
        usage='\n  mmpm env',
        help='display the MMPM environment variables and their value(s)'
    )

    # MMPM AND GLOBAL OPTIONS
    arg_parser.add_argument(
        '-v',
        '--version',
        action='store_true',
        help='display MMPM version number',
        dest='version'
    )

    arg_parser.add_argument(
        '--guided-setup',
        action='store_true',
        help='run the guided setup/installation of the GUI, environment variables, and other features',
        dest='guided_setup'
    )

    arg_parser.add_argument(
        '--migrate',
        action='store_true',
        help='migrate legacy file names and database keys. Only required once if prior version of MMPM is < 1.25',
        dest='migrate'
    )

    # hidden argument used for the GUI interface
    arg_parser.add_argument(
        '--GUI',
        action='store_true',
        default=False,
        help=argparse.SUPPRESS,
        dest='GUI'
    )

    argcomplete.autocomplete(arg_parser)

    if len(sys.argv) < 2:
        arg_parser.print_help()
        sys.exit(0)

    return arg_parser
